#include <SDL2/SDL.h>

class TB {
protected:

    SDL_Window*         sdl_window;
    SDL_Renderer*       sdl_renderer;
    SDL_Surface*        sdl_screen_surface;
    SDL_Texture*        sdl_screen_texture;
    SDL_Event           sdl_evt;
    Uint32*             sdl_screen_buffer;
    Uint32              pticks;

    Vppu* ppu;
    int x, y, _hs, _vs;

public:

    TB(int argc, char** argv);
    int     main();
    void    update();
    int     destroy();
    void    frame();
    void    pset(int x, int y, Uint32 cl);
    void    vga(int hs, int vs, int color);
};

TB::TB(int argc, char** argv) {

    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO)) {
        exit(1);
    }

    SDL_ClearError();
    sdl_window          = SDL_CreateWindow("NES", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, 1280, 960, SDL_WINDOW_SHOWN);
    sdl_renderer        = SDL_CreateRenderer(sdl_window, -1, SDL_RENDERER_PRESENTVSYNC);
    sdl_screen_buffer   = (Uint32*) malloc(640 * 480 * sizeof(Uint32));
    sdl_screen_texture  = SDL_CreateTexture(sdl_renderer, SDL_PIXELFORMAT_BGRA32, SDL_TEXTUREACCESS_STREAMING, 640, 480);
    SDL_SetTextureBlendMode(sdl_screen_texture, SDL_BLENDMODE_NONE);

    pticks = 0;
    ppu = new Vppu();

    x   = y   = 0;
    _hs = _vs = 1;
}

int TB::main() {

    for (;;) {

        Uint32 ticks = SDL_GetTicks();

        // Прием событий
        while (SDL_PollEvent(& sdl_evt))
        switch (sdl_evt.type) {
            case SDL_QUIT: return 0;
        }

        // Обновление экрана 60 FPS
        if (ticks - pticks >= 16) {
            pticks = ticks; update(); return 1;
        }

        SDL_Delay(1);
    }
}

void TB::update() {

    SDL_Rect dstRect;

    dstRect.x = 0;
    dstRect.y = 0;
    dstRect.w = 1280;
    dstRect.h = 960;

    SDL_UpdateTexture       (sdl_screen_texture, NULL, sdl_screen_buffer, 640 * sizeof(Uint32));
    SDL_SetRenderDrawColor  (sdl_renderer, 0, 0, 0, 0);
    SDL_RenderClear         (sdl_renderer);
    SDL_RenderCopy          (sdl_renderer, sdl_screen_texture, NULL, & dstRect);
    SDL_RenderPresent       (sdl_renderer);
}

int TB::destroy() {

    free(sdl_screen_buffer);
    SDL_DestroyTexture(sdl_screen_texture);
    SDL_DestroyRenderer(sdl_renderer);
    SDL_DestroyWindow(sdl_window);
    SDL_Quit();

    return 0;
}

void TB::pset(int x, int y, Uint32 cl) {

    if (x < 0 || y < 0 || x >= 640 || y >= 480)
        return;

    sdl_screen_buffer[640*y + x] = cl;
}

void TB::frame() {

    Uint32 timet = SDL_GetTicks();

    // 25 mhz / 60 fps ~ 415000
    for (int i = 0; i < 415000; i++) {

        ppu->clock = 0; ppu->eval();
        ppu->clock = 1; ppu->eval();

        vga(ppu->hs, ppu->vs, 65536*(ppu->r*16) + 256*(ppu->g*16) + (ppu->b*16));
    }

    timet = SDL_GetTicks() - timet;
}

// Экран VGA
void TB::vga(int hs, int vs, int color) {

    if (hs) x++;

    // Отслеживание изменений HS/VS
    if (_hs == 0 && hs == 1) { x = 0; y++; }
    if (_vs == 0 && vs == 1) { x = 0; y = 0; }

    // Вывод на экран
    pset(x - 49, y - 33, color);

    // Сохранить предыдущее значение
    _hs = hs;
    _vs = vs;
}
